(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_back.sml,v
   Comment: Back-end of CodGenC1. Translates intermediate representation to C.
   Release: scope-0-10 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   Does not actually dump the code, but builds everything what is needed in
   csystem record. Then a dumper actually solves the puzzle building real C
   files from snippets of code created here.

   CodGenC1_back.sml,v 1.64 2002/11/14 23:57:16 wasowski Exp 
*)

structure CodGenC1_back = struct 

(* some interface *)

open Err CodGenC1_LR CodGenC1_back_types CodGenC Say

infixr 1 $ 
infix  4 INIT
infix  3 --

structure SE = SymbolEnv
structure S  = Symbol

(* code generate simple integer typedef *)

fun cg_typedef (name :string) (size:int) :codgen =
	let val celltype = tyname $ uint_bysize $ size
	    val _ = sayln VHigh (name ^ ":" ^ itoa size ^ " " ^ celltype)
	in TYPEDEF celltype name end 
	handle e => intErr e 200209255 

val lrmnemonics = DEFINE "HIST" eoc 
			-- DEFINE "EVNT"	(sg "\t\t/* event reference */")
			-- DEFINE "ENEX(n)"	(sg "\t\t/* entry/exit ref  */")
			-- DEFINE "ACGD(n)"     (sg "\t\t/* action/guard    */")
			-- DEFINE "PCNC(n)"	(sg "\t\t/* pos/neg-cond    */")
			-- DEFINE "ORST"	(sg "\t\t/* or-state ref    */")
			-- DEFINE "ANDST"	(sg "\t\t/* and-state ref   */")

(* add named LR mnemonic to compund (encoded) value *)

fun compound_mnemonic (name :string) (ints :int list) :string list =
	if null ints then []
	else String.concat [name, "(", itoa (length ints), ") ", itoa (hd ints)]
		::mitoa (tl ints)

(* pack entry/exit actions. NOTE: this must be in synchro with corresponding
   measure_enex function in scheduler *)

fun pack_enex (sizes :lrsizes) (en :int,ex :int) :string list =
	let val ints = 
		if ex = null_actref then
			if en = null_actref then []	
			else pack1 en (#actionref sizes) (#orstref sizes)
		else pack2 (ex,en) (#actionref sizes,#actionref sizes) 
			(#orstref sizes)
	in compound_mnemonic "ENEX" ints end

fun cg_getexit (sizes :lrsizes) :codgen =
	let val _ = assert (fn () => (#actionref sizes <= #orstref sizes))
			"200210211 - unhandled type sizes combination"
			(* see comment below for acgd *)
	in	
		if (#actionref sizes < #orstref sizes) then
			DEFINE "GET_EXIT(s)"
				(cg ["((*(GET_AND(s)-1)) >> ",
				itoa (#actionref sizes),
				")"])
				(* if there is no action STMRK will do well!*)

		else (* #actionref siszes = #orstref sizes *)
			DEFINE "GET_EXIT(s)"
				(sg "((*(GET_AND(s)-1) == STMRK) ? \
					\STMRK : *(GET_AND(s)-2))")
	end

fun cg_getenter (sizes :lrsizes) :codgen = 
	let val _ = assert (fn () => (#actionref sizes <= #orstref sizes))
			"200210212 - unhandled type sizes combination"
			(* see comment below for acgd *)
	in	
		if (#actionref sizes < #orstref sizes) then
			DEFINE "GET_ENTER(s)"
				(cg ["((*(GET_AND(s)-1)) & ",
				itoa (ones (#orstref sizes) + 1),
				")"])
				(* if there is no action STMRK will do well!*)
		else (* #actionref siszes = #orstref sizes *)
			DEFINE "GET_ENTER(s)" (sg "(*(GET_AND(s)-1))")
	end

(* pack action and guard fields - used in transitions. NOTE: should be in
   synchro with measure_acgd function of scheduler *)

fun pack_acgd (sizes :lrsizes) (ac :int, gd :int) :string list =
	let val ints = 
		if gd = always_ref then							
			if ac = null_actref then []
			else pack1 ac (#actionref sizes) (#andstref sizes)
		else pack2 (ac,gd) (#actionref sizes,#guardref sizes) 
			(#andstref sizes)
	in compound_mnemonic "ACGD" ints end

(* FIXME: there is a fundmanental problem with runtime representation *)
(* action and guard become indistinguishable from HMRK/IMRK as soon   *)
(* as sizeof(guardref) > sizeof(andstref) or                          *)
(* sizeof(actionref) > sizeof(andstref). I leave the problem unsolved *)
(* for a moment, hoping that this never holds for models I have. It   *)
(* should however be fixed one day. For instance: save the youngest   *)
(* cell first and then exclude all values with youngest cell equal to *)
(* HMRK / IMRK / STMRK from being legal guardref and actionref        *)
(* Alternatively change internal representation (only in this case)   *)
(* explicitly adding compulosory field with number of bytes for action*)
(* and guard in total <- for each transition                          *)
(* BAD: to make things worse some part of this logics is also kept in *)
(* fire () (in CodGenC1c)                                             *)
(* FOR NOW I LEAVE A SANITY ASSERTION HERE			      *)

fun cg_unmangle_acgd guards (sizes :lrsizes) :codgen =
	let val _ = assert (fn () => (#actionref sizes <= #andstref sizes)
			andalso (#guardref sizes <= #andstref sizes))
			"200210210 - unhandled type sizes combination"
			(* see comment above *)
	in if not guards then
		DEFINE "UNMANGLE_AC(ptr,i)"
		(sg "{ (i) = *(ptr); if ( (i) >= MARKNO ) (ptr)++; }")

	else if (#actionref sizes + #guardref sizes <= #andstref sizes) then

			DEFINE "UNMANGLE_ACGD(ptr,i,j)"
			(sg "{ " -- access2 (#actionref sizes,#guardref sizes)
				(#andstref sizes) -- sg "{if ((j)==0) j = MARKNO;} }")
			
	else 
		DEFINE "UNMANGLE_ACGD(ptr,i,j)"
		(sg "{ (i) = *ptr; \
			\if (i >= MARKNO) (ptr)++; \
			\(j) = *ptr; \
			\if ( (j) >= MARKNO) (ptr)++;}")

	end

(* pack positive and negative counters *)

fun pack_pcnc (sizes :lrsizes) (pc :int, nc :int) :string list =
	let val ints = pack2 (pc,nc) (1,1) (#andstref sizes)
	in compound_mnemonic "PCNC" ints end

fun cg_unmangle_pcnc (sizes :lrsizes) :codgen =
	DEFINE "UNMANGLE_PCNC(ptr,i,j)" 
		(access2 (1,1) (#andstref sizes))

(* compile a LR instructions/fields *)

fun comp_marker STMRK   = "STMRK"
  | comp_marker HMRK    = "HMRK"
  | comp_marker IMRK    = "IMRK" 
  | comp_marker FPSCOP  = "FPSCOP"
  | comp_marker NFPSCOP = "NFPSCOP"

(* definitions are static and could be put in some static file *)
(* However this way we keep defines close to values used for   *)
(* compilation (above). Reducing the chance of desynchronizing *)

val markers =      DEFINE "STMRK"   (ig 0) 
		-- DEFINE "HMRK"    (ig 1) 
		-- DEFINE "IMRK"    (ig 2)
		-- DEFINE "FPSCOP"  (ig 3)
		-- DEFINE "NFPSCOP" (ig 4)
		-- DEFINE "MARKNO"  (ig 5)
		
(* compile a list of byte code entries. suitable to use in folds *)
(* FIXME: jakby to przerobic na codgen i sprawdzic czy predkosc  *)
(* sie zmienila?                                                 *)

fun comp_lr (sizes :lrsizes) ({hor_dict,hand_dict,eenv,...} :scheduling) 
	(atom :lratom,res :string list) : string list = (
	case atom of 
		  LRMrk mrk		=> comp_marker mrk :: res
		| LRAndStateRef s	=> "ANDST " ^ 
						itoa (lookup_stdict hand_dict s)
						::res
		| LROrStateRef  s	=> ("ORST " ^ 
						itoa (lookup_stdict hor_dict s)
						::res 
						handle e => intErr e 200209259)
		| LRNoEvent		=> "NOEVENT" ::res
		| LREventRef e		=> "EVNT " ^ itoa (lookup eenv e):: res
		| LRHistoryRef i	=> "HIST " ^ itoa i :: res
		| LREnEx (en,ex)	=> pack_enex sizes (en,ex) @ res
		| LRAcGd (ac,gd)	=> pack_acgd sizes (ac,gd) @ res
		| LRCondCounters(pc,nc)	=> pack_pcnc sizes (pc,nc) @ res
		| LRStateLabel _	=> res
		| LREventLabel _	=> res )
	handle e => intErr e 200209257

(* compile an array in LR *)

fun cg_arraylr (name :string) (tyname :string) (arr :lratom list) 
	(sizes :lrsizes) (sch :scheduling) :codgen =
	let val lst      = foldr (comp_lr sizes sch) [] arr
	    val iniblock = mk_commented_arrini lst 5 
	    val asize    = itoa (length lst)
	in ARRAY_DECL tyname name asize INIT iniblock end
	handle e => intErr e 200209256

(* compile signal groups. return signals array (option) and #define *)
(* SCOPE_SLIST for                                                  *)

fun cg_signals (signals :signals) (sizes :lrsizes) (sch :scheduling) 
						:codgen option * codgen =
	if UList.null signals then (NONE, UNDEF "SCOPE_SLIST")
	else let val arr = rev $ List.concat $ UList.toList signals
		 val def = cg_arraylr "signals" "const eventtag" arr sizes sch
	     in (SOME def, DEFINE "SCOPE_SLIST" eoc) end

(* compile history state *)

fun cg_history (history :lratom list) (sizes :lrsizes) (sch :scheduling)
							:codgen =
	if history = [LRMrk STMRK] then 
		UNDEF "SCOPE_HISTORY"  (* no history saved *)
	else cg_arraylr "history" "andstref" (rev history) sizes sch
		-- DEFINE "SCOPE_HISTORY" eoc

(* code generate access macro: macro which given stateid returns a pointer
   to the main cell of state representation *)

fun cg_stacc (dict :stdict) (dname :string) (hname :string) 
		(macro :string) :codgen =
	case dict of
		  Direct _	=> DEFINE (macro^"(d)") 
					(cg ["(",hname," + (d))"])
		| Indirect _	=> DEFINE (macro^"(d)") 
					(cg ["(",hname," + ", dname,"[d]"])

(* code-generate hierarchy access dictionaries if needed *)

fun cg_stdict (sizes :lrsizes) (dname :string) (dict :stdict) :codgen option =
	case dict of 
		  Direct _ => NONE (* no code generation needed *)
		| Indirect (mapping,lst,size) =>
			let val cty = "const " ^ (tyname $ uint_bysize 
							$ mintysize $ size-1)
			    val iniblock = mk_commented_arrini (mitoa lst) 10
			    val asize = itoa (length lst)
			in SOME (ARRAY_DECL cty dname asize INIT iniblock) end

(* always return the macro for accessing states          *)

fun cg_getst (dict :stdict) (macroname, arrayname, dictname) :codgen =
	case dict of
		  Direct _ => DEFINE (macroname ^ "(s)") 
			(cg ["(", arrayname, "+ (s))"])
		| Indirect _ => DEFINE (macroname ^ "(s)")
			(cg ["(", arrayname, "+ ", dictname, "[(s)] )"])

(* code generate symbolic dictionaries if needed *)

fun cg_symdict (debug :bool) (dname :string) (dict :stdict) :codgen option =
	if not debug then NONE
	else let fun mkstruct (a,b) = 
			String.concat ["{",itoa b,",",quote (S.name a),"}"]
		 val d    = case dict of Direct d => d 
				       | Indirect (d,_,_) => d
		 val null = mkstruct (S.symbol "ERROR",0)
		 val lst  = foldr (fn (x,e) => mkstruct x :: e) 
				[null] (SE.toList d) 
		 val inib = mk_commented_arrini lst 1
		 val asize = itoa (length lst)
	in SOME (ARRAY_DECL "const symbol" dname asize INIT inib) end

(* model independent symbolic dictionary entry type *)

val struct_symbol = sg "struct struct_symbol { int id; const char * name; };\n\
		       \typedef struct struct_symbol symbol;\n"

(* code generate name macros for events (only for external) *)

fun cg_evmacros (eenv :eenv) (evtypes :evtypes) :codgen =
	let val lst      = SE.toList evtypes 
	    val ext      = List.filter(fn (ev,(i,ty)) => not i) lst
	    val names    = map #1 ext
	    fun def name = DEFINE (Symbol.name name) 
				(sg $ itoa $ SE.lookup eenv name)
	in foldl (fn (na,e) => e -- def na) eoc names end

(* code generate symbolic names of events for debugging purposes *)
(* would be more efficient to do it in cg_evmacros, but this way *)
(* is cleaner.							 *)

fun cg_evdict debug (eenv :eenv) :codgen option =
	if not debug then NONE
	else let fun compare ((_,a),(_,b)) = Int.compare (a,b)
		 val lst  = Listsort.sort compare (SE.toList eenv)
		 val lst1 = map (quote o Symbol.name o #1) lst 
		 val inib = mk_commented_arrini lst1 1
		 val asize= itoa (length lst1)
	in SOME (ARRAY_DECL "const char *" "evdict" asize INIT inib) end

(* generate structure for single event type                  *)
(* FIXME: currently empty structures are generated for most  *)
(* of events (as most of events have no parameters). The code*)
(* would be more readable.				     *)

fun cg_evtype (_,(_,[]))			= eoc
  | cg_evtype (name :symbol,et :evtype) :codgen =
	let val tynames = map tyname (#2 et)
	    fun g (tn,(i,e)) : int * string list = 
		(i+1, "\t" ^ tn ^ " f" ^ itoa i ^ ";\n" ::e)
	in 
		cg ["\nstruct Struct", Symbol.name name, " {\n"]
		-- cg (#2 $ foldr g (0,[]) tynames)
		-- sg "};\n"
	end

(* generate structures for all event types *)

fun cg_evtypes (evtypes :evtypes) :codgen =
	let val lst = SE.toList evtypes
	in foldl (fn (et,e) => e -- cg_evtype et) eoc lst end

(* generate a union type of all events. Would more efficient *)
(* (but less clear) to do it in cg_evtypes.                  *)

fun cg_evunion (evtypes :evtypes) :codgen =
	let val lst = SE.toList evtypes
	    fun k ((n,(_,[])),e) = e
              | k ((n,(_,_)),(os,_)) = 
		(String.concat ["\t\tstruct Struct", Symbol.name n, 
			" _E_", Symbol.name n, ";\n"] ::os, true)
	    val (fields, nonempty) = foldl k ([],false) lst
	    val union = if nonempty then 
				sg "\tunion {\n" -- cg fields -- sg "\t} fields;\n"
			else eoc
	in sg "\nstruct StructEvent {\n\
			\\n\teventtag tag;\n" -- union -- sg "\n};\n\n"
		-- sg "\nextern struct StructEvent CurrEvent;\n"
	end
		    
(* generate transition table. I cannot use cg_arraylr because I need to know  *)
(* the size of trans array for type of traniter. Retvalues are: the array     *)
(* code and number of cells in the array                                      *)

fun cg_trans (sizes :lrsizes) (sch :scheduling) (tenv :tenv) :codgen * int =
	let val lst  = List.concat $ map #2 (SE.toList tenv)
	    val ilst = foldr (comp_lr sizes sch) [] lst
	    val iniblock = mk_commented_arrini ilst 5
	    val asize = itoa (length ilst)
	in (DEFINE "TRANS_MAX" (sg asize) -- eol -- eol
		-- ARRAY_DECL "const transcell" "trans" "TRANS_MAX" 
			INIT iniblock,
	    length ilst)
	end handle e => intErr e 200209252 

(* and transition index *)

fun cg_tranidx (sch :scheduling) (sizes :lrsizes) (tranlen :int) :codgen * codgen =
	let fun mkoffsets (ev,i) = (lookup (#eenv sch) ev,i)
	    fun compare ((a,_),(b,_)) = Int.compare (a,b)
	    val lst1 = (#evnumber sizes, tranlen)
			:: map mkoffsets (SE.toList (#tranidx sch))
	    val lst2 = map #2 $ Listsort.sort compare lst1 
	    val iniblock = mk_commented_arrini (mitoa lst2) 10
	    val asize = itoa (length lst2)
	in ( ARRAY_DECL "const tranref" "tranidx" asize INIT iniblock, 
	     cg_typedef "tranref" (mintysize $ List.last lst2)) 
	end
	    (* the last should be max... YUCK: this is slow *)
	    
(* generate variable definition (and extern decl for external variable) *)

fun cg_vardecl (id:symbol) (static :bool) (ty :ty) (ini :value) 
							:codgen * codgen =
	let fun f id (SimpleType ty  ) = tyname ty ^ " " ^ Symbol.name id
	      | f id (ArrayType(ty,n)) = tyname ty ^ " " ^ Symbol.name id 
					     ^ "[" ^ itoa n ^ "]"
	    val dcl = f id ty
	    val def = dcl ^ " = " ^ dump_value ini 
	in if static then (cg ["static ",def, ";\n"], eoc)
	   else (cg [def,";\n"], cg ["extern ",dcl,";\n"])
	end

fun cg_vars  (venv :venv) (vartypes :vartypes) :codgen * codgen =
	let val tlst = SE.toList vartypes
	    fun f (v,e) = let val ini = lookup venv (#1 v)
			      val internal :bool = #1 (#2 v)
			      val ty :ty = #2 (#2 v)
			      val (def,dcl) = cg_vardecl (#1 v) internal ty ini
			  in (#1 e -- def, #2 e --dcl ) end
	in foldl f (eoc,eoc) tlst end

(* generate constant definitions (preprocessor) *)

fun cg_consts (cenv :const env) :codgen =
	let val lst = map (fn (na,va) => (S.name na,dump_const va)) 
				(SE.toList cenv)
	    val cgs = map (fn (na,va) => DEFINE na (sg va)) lst 
	in foldr op-- eoc cgs end 

(* code generate expression *)	

fun cg_exp (e :aexp) :codgen =
	case e of
		  Cst c => sg (dump_const c)
		| Access a => cg_acc a 
		| Prim1  (ope,e1) =>
			cg [dump_Prim1 ope, "("] -- cg_exp e1 -- sg ")"
		| Prim2  (ope,e1,e2) =>
			sg "(" -- cg_exp e1 -- sg ")"
				-- sg (dump_Prim2 ope) --
			sg "(" -- cg_exp e2 -- sg ")"
		| Call oe => cg_oexp oe

(* code generate accesses - with substitution *)
		
and cg_acc (a :access) :codgen = 
	case a of AccessVar v => sg (Symbol.name v)
		| AccessArr (a,e) => 
			sg (Symbol.name a) -- sg "[" --  cg_exp e -- sg "]" 

(* generate output expressions - with substitution *)			
  
and cg_oexp ((f,[])  :oexp) :codgen = cg [Symbol.name f, "()"]
  | cg_oexp ((f,[p]) :oexp) :codgen = 
	cg [Symbol.name f, "("] -- cg_exp p -- sg ")"
  | cg_oexp ((f,h::t):oexp) :codgen =
	let val ch = cg_exp h
	    val ct = map cg_exp t
	in cg [Symbol.name f, "("] -- foldl (fn (f,s) => f -- sg "," -- s) 
		ch ct -- sg ")" 
	end

(* CG boolean expression *)

val cg_bexp = cg_exp

(* CG assignments *)

fun cg_assgn ((a,e) :assgn) :codgen =
	cg_acc a -- sg " = " -- cg_exp e 

(* code generate a guard. Guard should never be null as front end allocates   *)
(* "always" guard for all empty entries. Even this gurad is redundant in data *)
(* layouts with optional guards but in some layaouts (with packed guards) we  *)
(* need the value of "always"                                                 *)

fun cg_guard (g :guard) : codgen =
	let val ret = RETURN $ cg_bexp 
			$ foldl (fn (r,l) => Absyn.Prim2(Absyn.OpAnd,l,r)) 
			(hd g) (tl g)
	in	
		sg "\t\t" -- ret -- sg "\n" 
		before assert (fn () => not (null g))
	end

(* code generate a guard dispatcher *)

fun cg_guards (guards :guards) : codgen option * codgen =
	if UList.length guards <= 1 then (* only always *)
		(NONE, UNDEF "SCOPE_GUARDS") 
	else let val lst    = UList.toList guards
		 val lst1   = map cg_guard lst 
		 val ids    = List.tabulate (length lst, fn x => CASE (itoa (x+markno)))
		 val cases1 = ListPair.map (op $) (ids, lst1)
		 val cases2 = foldl (fn (x,e) => e -- sg "\t" -- x) eoc cases1
		 val body   = SWITCH "g" cases2 -- sg "\n\t\t return 1;"
		 val disp   = SOME (FUNC (sg "int eval ( const guardref g )") body)
		 val macro  = DEFINE "SCOPE_GUARDS" eoc
	in (disp, macro) end

(* code generate a signal action reference *)

fun cg_sigaction (eenv :eenv) (evnumber :int) (sigs :signals) 
					(signal  :Absyn.sexp list) :codgen =
	case signal of 
		  []  => eoc 
		| [(s,_)] => sg "\t\tSET_RETURN_SIGNAL("
				-- ig (SE.lookup eenv s) -- sg ");\n"
		(* a HORRIBLE repetition, have to recompile signals to 
		   check the address!!! FIXME *)
		| ss  => let val css = LRNoEvent::map (LREventRef o #1) signal
			     val vssignal = cg [itoa $ UList.find (css,sigs), 
						" + EVENTNO + 1"]
			 in sg "\t\tSET_RETURN_SIGNAL(" -- vssignal -- sg ");\n"
			 end handle UList.NotFound => impossible "#200209250"
	
(* code generate actions *)

fun cg_action (eenv :eenv) (evnumber :int) (sigs :signals)	
				{assgns,outputs,signals} :codgen = 
	let val cassgns  = foldl (fn (a,e) 
				=> e -- sg "\t\t" -- cg_assgn a -- sg ";\n") 
				eoc assgns
	    val outputs  = foldl (fn (ou,e) 
				=> e -- sg "\t\t" -- cg_oexp ou -- sg ";\n") 
				eoc outputs
	    val vssignal = cg_sigaction eenv evnumber sigs signals
	in  cassgns -- outputs -- vssignal -- sg "\t\tbreak;\n" end

(* Code generate action dispatcher *)

fun cg_actions (eenv :eenv) (evnumber :int) (sigs :signals) 
					(actions :actions) :codgen =
	let val lst    = rev (UList.toList actions)
	    val lst1   = map (cg_action eenv evnumber sigs) lst 
	    val ids    = List.tabulate (length lst,fn x => CASE (itoa (x+markno)))
	    val cases1 = ListPair.map (op$) (ids, lst1)
	    val cases2 = foldl (fn (x,e) => e -- sg "\t" -- x) eoc cases1
	    val cases3 = List.tabulate(markno, 
				fn i => sg "\t" -- CASE (itoa i) eoc)
	    val cases4 = CodGen.concat cases3 -- cases2
	in
		sg "void exec ( const actionref a ) { \n\				
		   \\t#ifdef SCOPE_SIGNALQUEUE\n\
			\\tsignalref _vssignal = NOEVENT;\n\
			\\textern void enqueue1 ( eventtag );\n\
			\#ifdef SCOPE_SLIST\n\
			\\textern void enqueueList ( signalref );\n\
			\#endif\n\
		   \\t#endif\n\n\t" --
		SWITCH "a" cases4 --
		sg "\n\t#ifdef SCOPE_SIGNALQUEUE\n\
			\\t#ifdef SCOPE_SLIST\n\
			\\tif (_vssignal < NOEVENT) enqueue1(_vssignal);\n\
			\\telse if (_vssignal > NOEVENT)\n\
			\\t\tenqueueList (_vssignal-NOEVENT-1);\n\
			\\t#else\n\
			\\tif (_vssignal != NOEVENT) enqueue1(_vssignal);\n\
			\\t#endif\n\
		   \\t#endif\n\
		   \}\n"
	end

fun cg_conf (name :string) (width :int) :codgen =
	let val iniblock = mk_commented_arrini (List.tabulate 
				(width, fn _ => "STMRK")) 10
	in VAR_DECL "static conf" name INIT iniblock end 

(* Cod generate signal queue #defines and array if needed *)

fun cg_queue (size :int option) :codgen =
	case size of 
		  NONE   => UNDEF "SCOPE_SIGNALQUEUE" 
				(* SET_RETURN_SIGNAL should never be needed *)
				(* in this case                             *)
		| SOME n => 
			let val lst      = List.tabulate (n,fn _ =>"NOEVENT")
			    val iniblock = mk_commented_arrini lst 7
			in 
				DEFINE "SCOPE_SIGNALQUEUE" eoc
				-- DEFINE "QUEUE_LENGTH" (ig n)
				-- ARRAY_DECL "static eventtag" "queue" "QUEUE_LENGTH"
					INIT iniblock 
				-- DEFINE "SET_RETURN_SIGNAL(s)" 
					(sg "{_vssignal = (s);}")
			end

(* #defines for PURE_EXIT and IMPURE_EXIT *)

fun cg_exitmode {pure,impure} :codgen =
	(if pure then DEFINE "PURE_EXIT" eoc 
	 else UNDEF "PURE_EXIT" )
	-- (if impure then DEFINE "IMPURE_EXIT"  eoc
	    else UNDEF "IMPURE_EXIT" )

fun cg_externhdr (sysname :string) (externhdr :string option) :codgen =
	let val hdr = case externhdr of NONE => sysname ^ "_extern.h" 
				      | SOME s => s
	in INCLUDE hdr end

(* Code generate system *)

fun cg_system (debug,externhdr) (sys :lrsystem) (sizes :lrsizes) 
					(sch :scheduling) : csystem =
	let val (vardefs,vardecls) = cg_vars (#venv sys) (#vartypes sys)
					handle e => intErr e 200209251
	    val (trans,tranlen)    = cg_trans sizes sch (#tenv sys)
	    val (tranidx,tranref)  = cg_tranidx sch sizes tranlen
					handle e => intErr e 200209253
	    val (guards,scopeguards)= cg_guards  (#guards sys)
	    val (signals,scopeslist)= cg_signals (#signals sys) sizes sch
	in { name        = #name sys,
	     lrmnemonics = lrmnemonics,
	     hor         = cg_arraylr "hor"  "const horcell"  (#hor sys)  sizes sch,
	     hand        = cg_arraylr "hand" "const handcell" (#hand sys) sizes sch,
	     hor_dict    = cg_stdict sizes "hor_dict"    (#hor_dict sch),
	     hand_dict   = cg_stdict sizes "hand_dict" (#hand_dict sch),
	     hor_acc     = cg_stacc (#hor_dict sch) "hor_dict" "hor" "ORST",
	     hand_acc    = cg_stacc (#hand_dict sch) "hand_dict" "hand" "ANDST",
	     and_syms    = cg_symdict debug "and_syms" (#hand_dict sch),
	     or_syms     = cg_symdict debug "or_syms"  (#hor_dict  sch),

	     symbol      = if debug then SOME struct_symbol else NONE,

	     trans       = trans,
	     tranidx     = tranidx,
	     
	     evmacros    = cg_evmacros (#eenv sch) (#evtypes sys),
	     evdict      = cg_evdict debug (#eenv sch),
	     evtypes     = cg_evtypes (#evtypes sys),
	     evunion     = cg_evunion (#evtypes sys),
	     
	     vardefs     = vardefs,
	     vardecls    = vardecls,
	     consts      = cg_consts (#cenv sys),
	     
	     actions     = cg_actions (#eenv sch) (#evnumber sizes) 
					(#signals sys) (#actions sys),
	     guards      = guards,
	     externhdr   = cg_externhdr (#name sys) externhdr,
	     signals     = signals,
	     history     = cg_history (#history sys) sizes sch,

	     ROOT_STATE  = DEFINE "ROOT_STATE"  (sg $ itoa $ 
				lookup_stdict(#hand_dict sch) $ #root sys),
	     STATE_WIDTH = DEFINE "STATE_WIDTH" (sg $ itoa $ #statew sys),
	     TGT_WIDTH   = DEFINE "TGT_WIDTH"   (sg $ itoa $ #targetw sys),
	     EVENTNO     = DEFINE "EVENTNO"     (sg $ itoa $ #evnumber sizes)
				-- DEFINE "NOEVENT" (sg "EVENTNO"),
	     queue       = cg_queue (#queuelen sys),
	     markers     = markers,

	     UNMANGLE_PCNC
			 = cg_unmangle_pcnc sizes,
	     UNMANGLE_ACGD
			 = cg_unmangle_acgd (UList.length (#guards sys) > 1) sizes,
	     GET_OR	 = cg_getst (#hor_dict sch) ("GET_OR","hor","hor_dict"),
	     GET_AND	 = cg_getst (#hand_dict sch)("GET_AND","hand","hand_dict"),
	     GET_ENTER   = cg_getenter sizes,
	     GET_EXIT    = cg_getexit sizes,

	     SCOPE_GUARDS= scopeguards,
	     SCOPE_SLIST = scopeslist,

	     exitmode    = cg_exitmode (#exitmode sys),

	     prev_conf   = cg_conf "prev_conf" (#statew sys),
	     next_conf   = cg_conf "next_conf" (#statew sys),
	     targets     = cg_conf "targets"   (#targetw sys),
	     
	     actionref   = cg_typedef "actionref" (#actionref sizes),
	     guardref    = cg_typedef "guardref"  (#guardref  sizes),
	     orstref     = cg_typedef "orstref"   (#orstref   sizes),
	     andstref    = cg_typedef "andstref"  (#andstref  sizes),
	     signalref   = cg_typedef "signalref" (#signalref sizes),
	     eventtag    = cg_typedef "eventtag"  (mintysize $ #evnumber sizes),
	     traniter    = cg_typedef "traniter"  (mintysize $ tranlen),
	     tranref     = tranref,

	     confiter    = cg_typedef "confiter"  (mintysize $ #statew sys),
	     conf        = TYPEDEF "andstref" ("conf["^(itoa $ #statew sys)^"]"),

	     tgtsiter    = cg_typedef "tgtsiter"  (mintysize $ #targetw sys),
	     queueiter   = case #queuelen sys of 
				  NONE => NONE
				| SOME n => SOME(cg_typedef "queueiter" (mintysize n)),
	     
	     (* model-independent typedefs, included here to avoid *)
	     (* creating more static files                         *)

	     types       = TYPEDEF "andstref" "horcell" -- eol
				-- TYPEDEF "orstref"  "handcell" -- eol
				-- TYPEDEF "andstref" "transcell" -- eol 
				-- TYPEDEF "andstref" "historyref" -- eol
	     }
	end handle e => intErr e 200209254

end
