(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/lib/Fifo.sig,v
   Comment: Applicative FIFO queues from SMLNJ library.

   Modified by Andrzej Wasowski for MoscowML 2.0

   Fifo.sig,v 1.1 2002/05/29 21:41:10 wasowski Exp
*)
(* COPYRIGHT (c)1993 by AT&T Bell Laboratories. *)

signature Fifo = sig

    type 'a fifo

    exception Dequeue

    val empty : 'a fifo
    val isEmpty : 'a fifo -> bool
    val enqueue : 'a fifo * 'a -> 'a fifo
    val dequeue : 'a fifo -> 'a fifo * 'a			(* Dequeue *)
    val delete : ('a fifo * ('a -> bool)) -> 'a fifo
    val head : 'a fifo -> 'a					(* Dequeue *)
    val peek : 'a fifo -> 'a option
    val length : 'a fifo -> int
    val contents : 'a fifo -> 'a list
    val app : ('a -> unit) -> 'a fifo -> unit
    val map : ('a -> 'b) -> 'a fifo -> 'b fifo
    val foldl : ('a * 'b -> 'b) -> 'b -> 'a fifo -> 'b
    val foldr : ('a * 'b -> 'b) -> 'b -> 'a fifo -> 'b

(*
	['a fifo] type of applicative fifo queues of elements of type 'a.

	[empty] empty queue.

	[isEmpty q] is true iff q is an empty queue (q=empty).

	[enqueue(q,x)] adds x to the end of queue q. Operates in constant
	time (worst-case).

	[dequeue(q,x)] returns the front element of queue q and the tail of
	the queue. Raises Dequeue if q is empty. Takes O(1) amortized time.

	[delete(q,p)] deletes all elements of q which fulfill predicate p.

	[head q] returns front element of the queue. Raises Dequeue if queue
	is empty. Takes O(1) amortized 
	
	[peek q] returns (SOME x) if x is the front element of queue.
	If the queue is empty returns NONE. Takes O(1) amortized time.

	[length q] returns the number of elements in the queue.

	[contents q] returns the list of elements of the queue. The front
	element of the queue is the front element of the list.

	[app f q] applies f consecutively to elements of q starting with
	head.

	[map f q] applies f consecutively to each element x of q starting with 
	head, returns the queue of results.

	[foldl op% e q] evaluates xn % (x(n-1) % ( ... % (x2 % (x1 % e))))
	where x1, x2, ..., x(n-1), xn are queue elements, x1 is the head, 
	and % is taken to be infixed.

	[foldr op% e q] evaluates x1 % (x2 % (... % (x(n-1) % (xn % e))...))
	where x1, x2, ..., x(n-1), xn are queue elements, x is the head 
	and % is taken to be infixed.
*)
  end (* FIFO *)
